from enum import Enum
import torch
from part.Utils import GraphData
from typing import Optional
import torch.distributed as dist
import DistCustomPool
from shared_graph_memory import GraphInfoInShm
import distparser as parser

class DistGraphData(GraphData):
    def __init__(self, path):
        self.rank = parser._get_worker_rank()
        path = path + '/rank_' + str(self.rank)
        print('load graph ',path)
        super(DistGraphData,self).__init__(path)


class DistributedDataLoader:
    ''' 
     Args:
            data_path: the path of loaded graph ,each part 0 of graph is saved on $path$/rank_0
            num_replicas: the num of worker
            

    '''
    def __init__(
            self,
            graph_name,
            graph,
            data_index_mask = None,
            sampler = None,
            collate_fn = None,
            batch_size: Optional[int]=None,
            shuffle:bool = True,
            seed:int = 0,
            drop_last = False,
            **kwargs
    ):
        assert sampler is not None
        self.pool = DistCustomPool.get_sampler_pool()
        self.graph_name = graph_name
        self.batch_size = batch_size
        self.num_workers = parser._get_num_sampler()
        self.queue_size = parser._get_queue_size()
        #if(queue_size is None):
        #    queue_size = self.num_workers * 4 if self.num_workers > 0 else 4
        self.num_pending = 0
        self.collate_fn = collate_fn
        self.current_pos = 0
        self.drop_last = drop_last
        self.recv_idxs = 0
        self.queue = []
        self.shuffle = shuffle
        self.is_closed = False
        self.sampler = sampler
        self.epoch = 0
        self.graph = graph
        
        self.graph_inshm = GraphInfoInShm(graph)
        self.sampler_info = sampler#._get_sample_info()
        #self.graph_inshm._copy_sampler_to_shame(*sampler._get_neighbors_and_deg())

        self.shuffle=shuffle
        self.seed=seed
        self.kwargs=kwargs
        self.rank = parser._get_worker_rank()
        self.data = torch.arange(start = graph.partptr[self.rank],end = graph.partptr[self.rank+1])
        if(data_index_mask is not None):
            self.data = torch.masked_select(self.data,data_index_mask)
        self.expected_idx = len(self.data) // self.batch_size
        if(not self.drop_last and len(self.data) % self.batch_size != 0):
            self.expected_idx += 1
        self.pool.set_collate_fn(self.collate_fn,self.sampler_info,self.graph_name,self.graph_inshm)
        dist.barrier()
        
    def __next__(self):
        #print(self.sampler,self.num_samples,self.kwargs)
        #self.sampleGraph(self.sampler,self.num_samplers,self.kwargs)
        #return self.sampleGraph(self.sampler,self.num_samples,**self.kwargs)
        num_reqs = self.queue_size - self.num_pending
        for _ in range(num_reqs):
            self._submit_request_next()
        if self.recv_idxs < self.expected_idx:
            result = self._get_data_from_queue()
            self.recv_idxs += 1
            self.num_pending -=1
            return result
        else :
            assert self.num_pending == 0
            raise StopIteration
    
    def _get_data_from_queue(self,timeout=1800):
        ret = self.pool.get_result(self.graph_name,timeout = timeout)
        return ret
    

    def _submit_request_next(self):
        next_data = self._next_data()
        if next_data is None:
            return
        self.pool.submit_task(self.graph_name,next_data)
        self.num_pending += 1

    def _next_data(self):
        if self.current_pos >= len(self.data):
            return None
        next_idx = None
        if self.current_pos + self.batch_size > len(self.data):
            if self.drop_last:
                return None
            else:
                next_idx = self.data[self.current_pos:]
        else:
            next_idx = self.data[self.current_pos:self.current_pos + self.batch_size]
        self.current_pos += len(next_idx)
        return next_idx

        
    def __iter__(self):
        if self.shuffle:
            indx = torch.randperm(self.data.size(0))
            self.data = self.data[indx]
        self.recv_idxs = 0
        self.current_pos = 0
        self.num_pending = 0
        return self



    def __del__(self):
        if self.pool is not None:
            self.pool.delete_collate_fn(self.graph_name)

    def set_epoch(self,epoch):
        self.epoch = epoch 