import sys
from os.path import abspath, join, dirname
sys.path.insert(0, join(abspath(dirname(__file__))))
import math
import torch
import torch.multiprocessing as mp
from typing import Optional, Tuple
#import ctypes
#sample_lib = ctypes.cdll.LoadLibrary('Sample/sample_cores.cpython-37m-x86_64-linux-gnu.so')
from base import BaseSampler, NegativeSampling
from sample_cores import get_neighbors, neighbor_sample_from_nodes, heads_unique

class NeighborSampler(BaseSampler):
    def __init__(
        self,
        num_nodes: int,
        num_layers: int,
        fanout: list,
        workers = 1,        
        edge_index : torch.Tensor = None,
        deg = None,
        neighbors = None
    ) -> None:
        r"""__init__
        Args:
            edge_index: all edges in the graph
            num_nodes: the num of all nodes in the graph
            num_layers: the num of layers to be sampled
            fanout: the list of max neighbors' number chosen for each layer
            workers: the number of threads, default value is 1
        """
        super().__init__()
        self.num_layers = num_layers
        # 线程数不超过torch默认的omp线程数
        self.workers = min(workers, torch.get_num_threads())
        self.fanout = fanout
        self.num_nodes = num_nodes
        if(edge_index is not None):
            row, col = edge_index
            tnb = get_neighbors(row.tolist(), col.tolist(), num_nodes, self.workers)
            self.neighbors = tnb.neighbors
            self.deg = tnb.deg
        else:
            assert deg is not None
            assert neighbors is not None
            self.neighbors = neighbors
            self.deg = deg
    
    def _get_sample_info(self):
        return self.num_nodes,self.num_layers,self.fanout,self.workers
    
    def _get_neighbors_and_deg(self):
        return self.neighbors,self.deg
    
    def _set_neighbors_and_deg(self,neighbors,deg):
        self.neighbors = neighbors
        self.deg = deg
        
    def sample_from_nodes(
        self,
           nodes: torch.Tensor
    ) -> Tuple[torch.Tensor, list]:
        r"""Performs mutilayer sampling from the nodes specified in: nodes
        The specific number of layers is determined by parameter: num_layers
        returning a sampled subgraph in the specified output format: Tuple[torch.Tensor, list].

        Args:
            nodes: the list of seed nodes index
        Returns:
            sampled_nodes: the node sampled
            sampled_edge_index_list: the edge sampled
        """
        sampled_edge_index_list = []
        sampled_nodes = torch.IntTensor([])
        src_nodes = nodes.tolist()
        assert self.workers > 0, 'Workers should be positive integer!!!'
        for i in range(0, self.num_layers):
            sampled_nodes_i, sampled_edge_index_i = self._sample_one_layer_from_nodes(nodes, self.fanout[i])
            sampled_nodes = torch.cat([sampled_nodes, sampled_nodes_i])
            nodes = torch.unique(sampled_edge_index_i[1])
            sampled_edge_index_list.append(sampled_edge_index_i)
        sampled_nodes = heads_unique(sampled_nodes.tolist(), src_nodes)
        return torch.tensor(sampled_nodes), sampled_edge_index_list
    
    def sample_from_edges(
        self,
        edges: torch.Tensor,
        edge_label: Optional[torch.Tensor] = None,
        neg_sampling: Optional[NegativeSampling] = None
    ) -> Tuple[torch.Tensor, list]:
        r"""Performs sampling from the edges specified in :obj:`index`,
        returning a sampled subgraph in the specified output format.

        Args:
            edges: the list of seed edges index
            edge_label: the label for the seed edges.
            neg_sampling: The negative sampling configuration
        Returns:
            sampled_nodes: the nodes sampled
            sampled_edge_index_list: the edges sampled
            metadata: other infomation
        """
        src, dst = edges
        num_pos = src.numel()
        num_neg = 0
        if edge_label is None:
            edge_label = torch.ones(num_pos)

        if neg_sampling is not None:
            num_neg = math.ceil(num_pos * neg_sampling.amount)
            if neg_sampling.is_binary():
                src_neg = neg_sampling.sample(num_neg, self.num_nodes)
                src = torch.cat([src, src_neg], dim=0)
                dst_neg = neg_sampling.sample(num_neg, self.num_nodes)
                dst = torch.cat([dst, dst_neg], dim=0)
            if neg_sampling.is_triplet():
                dst_neg = neg_sampling.sample(num_neg, self.num_nodes)
                dst = torch.cat([dst, dst_neg], dim=0)

        seed = torch.cat([src, dst], dim=0)
        seed, inverse_seed = seed.unique(return_inverse=True)
        sampled_nodes, sampled_edge_index_list = self.sample_from_nodes(seed)

        if neg_sampling is None or neg_sampling.is_binary():
            edge_label_index = inverse_seed.view(2, -1)
            metadata = {'edge_label_index':edge_label_index, 'edge_label':edge_label}

        elif neg_sampling.is_triplet():
            src_index = inverse_seed[:num_pos]
            dst_pos_index = inverse_seed[num_pos:2 * num_pos]
            dst_neg_index = inverse_seed[2 * num_pos:]
            dst_neg_index = dst_neg_index.view(num_pos, -1).squeeze(-1)
            metadata = {'src_index':src_index, 'dst_pos_index':dst_pos_index, 'dst_neg_index':dst_neg_index}
        
        return sampled_nodes, sampled_edge_index_list, metadata

    def _sample_one_layer_from_nodes(
        self,
        nodes: torch.Tensor,
        fanout: int
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        r"""Performs sampling from the nodes specified in: nodes,
        returning a sampled subgraph in the specified output format: Tuple[torch.Tensor, torch.Tensor].

        Args:
            nodes: the list of seed nodes index
            fanout: the number of max neighbors chosen
        Returns:
            sampled_nodes: the nodes sampled
            sampled_edge_index: the edges sampled
        """
        tgb = neighbor_sample_from_nodes(nodes.tolist(), self.neighbors, self.deg, fanout, self.workers)
        row = torch.IntTensor(tgb.row())
        col = torch.IntTensor(tgb.col())
        sampled_nodes = torch.IntTensor(tgb.nodes())
        return sampled_nodes, torch.stack([row,col], dim=0)

    # def _sample_one_layer_from_nodes_parallel(
    #     self, 
    #     nodes: torch.Tensor,
    #     fanout: int
    # ) -> Tuple[torch.Tensor, torch.Tensor]:
    #     r"""Performs sampling from the nodes specified in: nodes,
    #     returning a sampled subgraph in the specified output format: Tuple[torch.Tensor, torch.Tensor].

    #     Args:
    #         nodes: the list of seed nodes index
    #         fanout: the number of max neighbor chosen
    #     Returns:
    #         sampled_nodes: the node sampled
    #         sampled_edge_index: the edge sampled
    #     """
    #     sampled_nodes=torch.IntTensor([])
    #     row=torch.IntTensor([])
    #     col=torch.IntTensor([])
    #     assert self.workers > 0, 'Workers should be positive integer!!!'
    #     with mp.Pool(processes=self.workers) as p:
    #         n=len(nodes)
    #         if(self.workers>=n):
    #             results = [p.apply_async(self._sample_one_layer_from_nodes, 
    #                                      (torch.tensor([node.item()]), fanout))
    #                        for node in nodes]
    #         else:
    #             quotient = n//self.workers
    #             remainder = n%self.workers
    #             # 每个batch先分配quotient个nodes，然后将余数remainder平均分配给其中一些batch
    #             nodes1 = nodes[0:(quotient+1)*(remainder)].resize_(remainder,quotient+1)# 分配了余数的batch
    #             nodes2 = nodes[(quotient+1)*(remainder):n].resize_(self.workers - remainder,quotient)# 未分配余数的batch
    #             results = [p.apply_async(self._sample_one_layer_from_nodes, 
    #                                     (nodes1[i], fanout)) 
    #                       for i in range(0, remainder)]
    #             results.extend([p.apply_async(self._sample_one_layer_from_nodes, 
    #                                          (nodes2[i], fanout)) 
    #                             for i in range(0, self.workers - remainder)])
    #         for result in results:
    #             sampled_nodes_i, sampled_edge_index_i = result.get()
    #             sampled_nodes = torch.unique(torch.cat([sampled_nodes, sampled_nodes_i]))
    #             row = torch.cat([row, sampled_edge_index_i[0]])
    #             col = torch.cat([col, sampled_edge_index_i[1]])
    #     return sampled_nodes, torch.stack([row, col], dim=0)

        # 不使用_sample_one_layer_from_node直接取所有点邻居方法：
        # row, col = edge_index
        # neighbors1=torch.concat([row[row==nodes[i]] for i in range(0, nodes.shape[0])])
        # neighbors2=torch.concat([col[row==nodes[i]] for i in range(0, nodes.shape[0])])
        # neighbors=torch.stack([neighbors1, neighbors2], dim=0)
        # print('neighbors: \n', neighbors)

if __name__=="__main__":
    edge_index1 = torch.tensor([[0, 1, 1, 1, 2, 2, 2, 4, 4, 4, 5], # , 3, 3
                                [1, 0, 2, 4, 1, 3, 0, 3, 5, 0, 2]])# , 2, 5
    num_nodes1 = 6
    num_neighbors = 2
    # Run the neighbor sampling
    sampler=NeighborSampler(edge_index=edge_index1, num_nodes=num_nodes1, num_layers=3, workers=4, fanout=[2, 1, 1])

    # neighbor_nodes, sampled_edge_index = sampler._sample_one_layer_from_nodes(nodes=torch.tensor([2,1]), fanout=num_neighbors)
    neighbor_nodes, sampled_edge_index = sampler.sample_from_nodes(torch.tensor([1,2]))
    
    # Print the result
    print('neighbor_nodes_id: \n',neighbor_nodes, '\nedge_index: \n',sampled_edge_index)
